#ifndef _HSV_HPP_
#define _HSV_HPP_

namespace histodet {

/**
 * Return min and max value among three values
 */
template<typename T>
inline void minmax3(T x, T y, T z, T &minv, T &maxv)
{
	T a, b;
	if (x >= y) { a = y; b = x; } else { a = x; b = y; }
	if (z >= a) minv = a; else minv = z;
	if (z >= b) maxv = z; else maxv = b;
}


/**
 * RGB to HSV conversion (float version)
 * @param r red [0,255]
 * @param g green [0,255]
 * @param b blue [0,255]
 * @param[out] h hue [0,360]
 * @param[out] s saturation [0,1]
 * @param[out] v value [0,255]
 * 
 * @param angle_origin is the hue angle for which you want hue=0, between 0 and 360
 */
template<int hue_max>
inline void rgb2hsv(float r, float g, float b, float &h, float &s, float &v, float angle_origin = 0.0f)
{
	float minv, maxv;
	minmax3(r, g, b, minv, maxv);
	float delta = maxv - minv;

	// h
	if (maxv == minv) h = 0.0f; else
	if (maxv == r) { h = (g-b)/delta; if (h < 0.0f) h += 6.0f; } else
	if (maxv == g) h = (b-r)/delta + 2.0f; else
		h = (r-g)/delta + 4.0f;
	h += angle_origin/60.0f; if (h >= 6.0f) h -= 6.0f;
	h *= ((float)hue_max)/6.0f;

	// s
	if (maxv == 0.0f) s = 0.0f; else s = delta / maxv;

	// v
	v = maxv;
}


/**
 * RGB to Hue conversion (uchar version)
 * @param r red [0,255]
 * @param g green [0,255]
 * @param b blue [0,255]
 * @return hue [0,255]
 */
inline
uchar rgb2h(uchar r, uchar g, uchar b)
{
	uchar h, minv, maxv;
	minmax3(r, g, b, minv, maxv);
	unsigned short delta = (maxv - minv)*2;

	// h
	if (maxv == minv) h = 0; else
	if (maxv == r) h = (b<=g) ? ((g-b)*85)/delta : 255-((b-g)*85)/delta; else
	if (maxv == g) h = 85 + ((b-r)*85)/delta; else
		h = 170 + ((r-g)*85)/delta;

	return h;
}


} // namespace histodet


#endif

