#ifndef _OBJECT_EXTRACTOR_HPP_
#define _OBJECT_EXTRACTOR_HPP_

#include <vector>
#include <queue>
#include <opencv2/core/core.hpp>

#include "histodet/object.hpp"

#ifdef CHRONO
#include <sys/time.h>
#include <iostream>
#endif

#if DISPLAY
#include <iomanip>
#include <opencv2/highgui/highgui.hpp>
#endif

namespace histodet {

/**
 * Abstract class for extracting objects in a thresholded detection result image.
 */
class ObjectExtractorAbstract
{
	protected:
		typedef cv::Mat PointList;
		PointList points;
		void extractPoints(cv::Mat const & detImage, float threshold, bool inverted, int window_size);
		inline bool isInRect(cv::Point const & point, cv::Rect const & rect)
			{ return point.x >= rect.x && point.x < rect.x+rect.width && point.y >= rect.y && point.y < rect.y+rect.height; }
	public:
		/**
		 * Extract the objects in an image
		 * @param detImage the detection result image
		 * @param threshold the threshold to apply
		 * @param inverted whether the distance is inverted or not
		 * @param window_size the detector window size that has been used
		 */
		virtual ObjectList extract(cv::Mat const & detImage, float threshold, bool inverted, int window_size) = 0;
};






/**
 * When searching only one instance of the object in the image, just finds
 * the largest blob if it is large enough.
 */
class ObjectExtractorBiggest: ObjectExtractorAbstract
{
	public:
		ObjectList extract(cv::Mat const & detImage, float threshold, bool inverted, int window_size);

};




/**
 * Use Kmeans clustering algorithm to extract several objects
 *
 */

class ObjectExtractorKmeans: ObjectExtractorAbstract
{
	public:
		ObjectList extract(cv::Mat const & detImage, float threshold, bool inverted, int window_size);

};



/**
 * Use flood fill algorithm to extract several objects faster
 *
 */



class ObjectExtractorFloodFill: ObjectExtractorAbstract
{
	private:
		template<typename T> class COMP_INF
			{ public: static inline bool comp(T const &a, T const &b) { return a <= b; } };
		template<typename T> class COMP_SUP
			{ public: static inline bool comp(T const &a, T const &b) { return a >= b; } };
		template<class COMP>
		ObjectList extract(cv::Mat const & detImage, float threshold, int window_size);
		
	public:
		ObjectList extract(cv::Mat const & detImage, float threshold, bool inverted, int window_size)
		{
			if (inverted) return extract<COMP_SUP<double> >(detImage, threshold, window_size);
			         else return extract<COMP_INF<double> >(detImage, threshold, window_size);
		}

};




///*********************************************************************************************************************************************
/// Template functions
///*********************************************************************************************************************************************



struct Blob2
{
	int xmin, xmax;
	int ymin, ymax;
	int n;
	Blob2(int x0, int y0): xmin(x0), xmax(x0), ymin(y0), ymax(y0), n(1) {}
};


template<class COMP>
ObjectList ObjectExtractorFloodFill::extract(cv::Mat const & detImage, float threshold, int window_size)
{
	#ifdef CHRONO
	static struct timeval tv1, tv2;
	static double avgt = 0., mint = 1e6;
	static int nchrono = 0;
	gettimeofday(&tv1,NULL);
	#endif

	cv::Mat pimg = cv::Mat(detImage.rows, detImage.cols, CV_8U, cv::Scalar(128)); // progress image: 128 is unknown, 255 is positive, 0 is negative

	std::vector<Blob2> blobs;
	std::queue<cv::Point> queue;
	cv::Point nb[4]; // neighbors

	// start by going through the detImage to find points above threshold
	for(unsigned i = window_size/2; i < detImage.rows-window_size/2; ++i)
		for(unsigned j = window_size/2; j < detImage.cols-window_size/2; ++j)
		{
			if (pimg.at<uchar>(i,j) == 128)
				if (COMP::comp(detImage.at<float>(i,j), threshold))
//				if (detImage.at<float>(i,j) <= threshold)
				{
					Blob2 blob(j,i);
					queue.push(cv::Point(j,i));
					pimg.at<uchar>(i,j) = 255;

					// start flood fill
					while (!queue.empty())
					{
						cv::Point & p = queue.front();

						if (p.y < blob.ymin) blob.ymin = p.y;
						if (p.y > blob.ymax) blob.ymax = p.y;
						if (p.x < blob.xmin) blob.xmin = p.x;
						if (p.x > blob.xmax) blob.xmax = p.x;
						blob.n++;

						nb[0].x = p.x-1; nb[0].y = p.y;
						nb[1].x = p.x+1; nb[1].y = p.y;
						nb[2].x = p.x; nb[2].y = p.y-1;
						nb[3].x = p.x; nb[3].y = p.y+1;
						for(int k = 0; k < 4; ++k)
							if (pimg.at<uchar>(nb[k].y,nb[k].x) == 128)
								if (COMP::comp(detImage.at<float>(nb[k].y,nb[k].x), threshold))
//								if (detImage.at<float>(nb[k].y,nb[k].x) <= threshold)
								{
									pimg.at<uchar>(nb[k].y,nb[k].x) = 255;
									queue.push(cv::Point(nb[k].x, nb[k].y));
								} else pimg.at<uchar>(nb[k].y,nb[k].x) = 0;

						queue.pop();
					}
					blobs.push_back(blob);
				} else pimg.at<uchar>(i,j) = 0;
		}

	ObjectList result;
	int min_points = window_size*window_size/4; // FIXME set this in histo file?
	int max_size = window_size*3;
	for(int i = 0; i < blobs.size(); ++i)
	{
		if (blobs[i].n > min_points && (blobs[i].xmax-blobs[i].xmin+1) < max_size && (blobs[i].ymax-blobs[i].ymin+1) < max_size)
		{
			Object obj(1);
			obj.set(Object::lpPosU, (blobs[i].xmax+blobs[i].xmin)/2, window_size / 3.0 / 3.0);
			obj.set(Object::lpPosV, (blobs[i].ymax+blobs[i].ymin)/2, window_size / 3.0 / 3.0);
			obj.set(Object::lpSize, window_size, window_size / 3.0 / 3.0);
			result.push_back(obj);
		}
	}

#ifdef CHRONO
	gettimeofday(&tv2,NULL);
	double t = (tv2.tv_sec-tv1.tv_sec) + (tv2.tv_usec-tv1.tv_usec)*1e-6;
	avgt += t; if (t < mint) mint = t; ++nchrono;
	std::cout << "[ObjectExtractor] Processed image in " << t << " seconds (average " << avgt/nchrono << " , min " << mint << ")" << std::endl;
#endif


#if DISPLAY
	static int nim = 0;
	std::ostringstream oss; oss << std::setw(2) << std::setfill('0') << nim;
	cv::Mat img_write, img_channels[3];
	for(int i = 0; i < 3; ++i) pimg.copyTo(img_channels[i]);
	cv::merge(img_channels, 3, img_write);
	for(int i = 0; i < blobs.size(); ++i)
		cv::rectangle(img_write, cv::Rect(blobs[i].xmin-1, blobs[i].ymin-1, blobs[i].xmax-blobs[i].xmin+3, blobs[i].ymax-blobs[i].ymin+3), cv::Scalar(0,255,0), 1);
	cv::imwrite(oss.str()+std::string("_clusters1.png"), img_write);
#endif

	return result;
}





} // namespace histodet



#endif
