
#include "histodet/bucketer.hpp"

namespace histodet {

	///***************************************************************************
	/// BucketerAbstract
	///***************************************************************************

	BucketerAbstract* BucketerAbstract::createBucketer(std::string hist_file, int & hist_size, int & hist_nbits)
	{
		cv::FileStorage fs(hist_file, cv::FileStorage::READ);
		if (!fs.isOpened()) { std::cerr << "BucketerAbstract::createBucketer: could not open " << hist_file << std::endl; return NULL; }

		cv::Mat tmp;
		fs["hist"] >> tmp;
		hist_size = tmp.rows;
		hist_nbits = 0; int s = hist_size/2; while (s != 0) ++hist_nbits, s/=2;

		double param1, param2;
		BucketerType bucketer_type;
		fs["bucketer_type"] >> (int&)bucketer_type;
		fs["bucketer_param1"] >> param1;
		fs["bucketer_param2"] >> param2;

		switch (bucketer_type)
		{
			case btGray:          return new BucketerGray(hist_nbits);
			case btHue:           return new BucketerHue<false>(1<<hist_nbits, hist_nbits+1);
			case btHueBruteforce: return new BucketerHueBruteforce(hist_nbits);
			case btHueLowsat:     return new BucketerHueLowsat<true, false>(1<<hist_nbits, hist_nbits+1, param1, param2);
			case btHueValue:      return new BucketerHueValue<false>(1<<(hist_nbits-1), 1<<(hist_nbits-1), hist_nbits);
			case btRGB:           return new BucketerRGB(hist_nbits);
			case btHueHighsat:     return new BucketerHueLowsat<false, false>(1<<hist_nbits, hist_nbits+1, param1, param2);
			default:              std::cerr << "BucketerAbstract::createBucketer: Unknown bucketer type " << bucketer_type << " in " << hist_file << std::endl; return NULL;
		}
	}


	///***************************************************************************
	/// BucketerLut
	///***************************************************************************

	template<typename T, bool color, bool mask>
	cv::Mat& BucketerLut<T,color,mask>::convertImage(cv::Mat const &img)
	{
		if (rgb_order) return convertImage_impl<0,1,2>(img);
		          else return convertImage_impl<2,1,0>(img);
	}
	
	template<typename T, bool color, bool mask>
	template<int idxR, int idxG, int idxB>
	cv::Mat& BucketerLut<T,color,mask>::convertImage_impl(cv::Mat const &img)
	{
		if (color != (img.channels()>=3)) std::cerr << "BucketerLut::convertImage: Image color does not match" << std::endl;
			
		// check bucketImage has right size
		if (img.cols != bucketImg.cols || img.rows != bucketImg.rows)
			bucketImg = cv::Mat::zeros(img.rows, img.cols, cv_type);

		// do the conversion
		if (color)
		{
			int channels = img.channels();
			for(int y = 0; y < img.rows; ++y)
			for(int x = 0; x < img.cols; ++x)
			{
				unsigned char *pix = img.data+img.step[0]*y+x*channels;
				*((T*)(bucketImg.data+bucketImg.step[0]*y)+x) =
					(mask && *(pix+idxR) == 0 && *(pix+idxG) == 0 && *(pix+idxB) == 0) ?
					-1	:
					lut[
					(*(pix+idxR) >> lut_tobin) * lut_size2 +
					(*(pix+idxG) >> lut_tobin) * lut_size  +
					(*(pix+idxB) >> lut_tobin)              ];
			}
		} else
		{
			// a LUT should not be necessary for gray images, except if we want complex mapping
			// we assume that there is no need to subsample the LUT in gray to avoid bit shift
			if (lut_size != 256) std::cerr << "BucketerLut::convertImage: LUT size for gray images must be 256 (8 bits)" << std::endl;
			for(int y = 0; y < img.rows; ++y)
			for(int x = 0; x < img.cols; ++x)
				*((T*)(bucketImg.data+bucketImg.step[0]*y)+x) = lut[*(img.data+img.step[0]*y+x)];
		}

		return bucketImg;
	}


	///***************************************************************************
	/// BucketerGray
	///***************************************************************************
	
	cv::Mat& BucketerGray::convertImage(cv::Mat const &img)
	{
		// check bucketImage has right size
		if (img.cols != bucketImg.cols || img.rows != bucketImg.rows)
			bucketImg = cv::Mat::zeros(img.rows, img.cols, CV_8U);
		
		// do the conversion
		for(int y = 0; y < img.rows; ++y)
		for(int x = 0; x < img.cols; ++x)
			bucketImg.ptr<unsigned char>(y)[x] = img.ptr<unsigned char>(y)[x] >> hist_tobin;
		//img.data+img.step[0]*y+x

		return bucketImg;
	}

	
	///***************************************************************************
	/// BucketerHue
	///***************************************************************************

	template<bool mask>
	BucketerHue<mask>::BucketerHue(int hist_size, int lut_nbits):
		BucketerLut<unsigned char, true, mask>(hist_size, lut_nbits)
	{
		float lut_bucket_size = 256.0 / this->lut_size;
		float hist_bucket_size = 360.0 / hist_size;
		float h, s, v;
		for(int i = 0; i < this->lut_size; ++i)
		for(int j = 0; j < this->lut_size; ++j)
		for(int k = 0; k < this->lut_size; ++k)
		{
			rgb2hsv<360>((i+0.5) * lut_bucket_size,
				(j+0.5) * lut_bucket_size,
				(k+0.5) * lut_bucket_size,
				h, s, v);
			this->lut[i * this->lut_size2 + j * this->lut_size + k] = (int)(h / hist_bucket_size);
		}
	}

	template class BucketerHue<false>;
	template class BucketerHue<true>;
	
	///***************************************************************************
	/// BucketerHueBruteforce
	///***************************************************************************

	cv::Mat& BucketerHueBruteforce::convertImage(cv::Mat const &img)
	{
		if (rgb_order) return convertImage_impl<0,1,2>(img);
		          else return convertImage_impl<2,1,0>(img);
	}

	template<int idxR, int idxG, int idxB>
	cv::Mat& BucketerHueBruteforce::convertImage_impl(cv::Mat const &img)
	{
		// check bucketImage has right size
		if (img.cols != bucketImg.cols || img.rows != bucketImg.rows)
			bucketImg = cv::Mat::zeros(img.rows, img.cols, CV_8U);

		// do the conversion
		float float_tobin = 360.0 / hist_size;
		int channels = img.channels();
		for(int y = 0; y < img.rows; ++y)
		for(int x = 0; x < img.cols; ++x)
		{
			unsigned char *pix = img.data+img.step[0]*y+x*channels;
			bucketImg.ptr<unsigned char>(y)[x] =
				rgb2h(*(pix+idxR), *(pix+idxG), *(pix+idxB)) >> hist_tobin;

// 			float h,s,v;
// 			rgb2hsv<360>(*(pix+2), *(pix+1), *(pix), h, s, v);
// 			*((unsigned short*)(bucketImg.data+bucketImg.step[0]*y)+x) =
// 				(int)(h / float_tobin);
		}
		
		return bucketImg;
	}

	
	///***************************************************************************
	/// BucketerHueLowsat
	///***************************************************************************

	template<bool includelowsat, bool mask>
	BucketerHueLowsat<includelowsat, mask>::BucketerHueLowsat(int hist_size, int lut_nbits, int minsat_, int minval_):
		BucketerLut<unsigned char, true, mask>(hist_size, lut_nbits), minsat(minsat_/255.), minval(minval_)
	{
		float lut_bucket_size = 256.0 / this->lut_size;
		float hist_bucket_size = 360.0 / (includelowsat ? (hist_size-1) : hist_size);
		float h, s, v;
		for(int i = 0; i < this->lut_size; ++i)
		for(int j = 0; j < this->lut_size; ++j)
		for(int k = 0; k < this->lut_size; ++k)
		{
			rgb2hsv<360>((i+0.5) * lut_bucket_size,
				(j+0.5) * lut_bucket_size,
				(k+0.5) * lut_bucket_size,
				h, s, v);
			this->lut[i * this->lut_size2 + j * this->lut_size + k] = (s < minsat || v < minval) ? (includelowsat ? hist_size-1 : -1) : (int)(h / hist_bucket_size);
		}
	}

	template class BucketerHueLowsat<false,false>;
	template class BucketerHueLowsat<false,true>;
	template class BucketerHueLowsat<true,false>;
	template class BucketerHueLowsat<true,true>;
	
	///***************************************************************************
	/// BucketerHueValue
	///***************************************************************************

	template<bool mask>
	BucketerHueValue<mask>::BucketerHueValue(int hist_hue_size, int hist_value_size, int lut_nbits):
		BucketerLut<unsigned char, true, mask>(hist_hue_size+hist_value_size, lut_nbits)
	{
		float lut_bucket_size = 256.0 / this->lut_size;
		float hist_hue_bucket_size = 360.0 / hist_hue_size;
		float hist_value_bucket_size = 256.0 / hist_value_size;
		float h, s, v;
		for(int i = 0; i < this->lut_size; ++i)
		for(int j = 0; j < this->lut_size; ++j)
		for(int k = 0; k < this->lut_size; ++k)
		{
			rgb2hsv<360>((i+0.5) * lut_bucket_size,
				(j+0.5) * lut_bucket_size,
				(k+0.5) * lut_bucket_size,
				h, s, v);
			if (s < 0.3 || v < 32)
				this->lut[i * this->lut_size2 + j * this->lut_size + k] = hist_hue_size + (int)(v / hist_value_bucket_size); else
				this->lut[i * this->lut_size2 + j * this->lut_size + k] = (int)(h / hist_hue_bucket_size);
		}
	}

	template class BucketerHueValue<false>;
	template class BucketerHueValue<true>;

	///***************************************************************************
	/// BucketerRGB
	///***************************************************************************

	cv::Mat& BucketerRGB::convertImage(cv::Mat const &img)
	{
		if (rgb_order) return convertImage_impl<0,1,2>(img);
		          else return convertImage_impl<2,1,0>(img);
	}

	template<int idxR, int idxG, int idxB>
	cv::Mat& BucketerRGB::convertImage_impl(cv::Mat const &img)
	{
		// check bucketImage has right size
		if (img.cols != bucketImg.cols || img.rows != bucketImg.rows)
			bucketImg = cv::Mat::zeros(img.rows, img.cols, CV_16U);

		// do the conversion
		int channels = img.channels();
		for(int y = 0; y < img.rows; ++y)
		for(int x = 0; x < img.cols; ++x)
		{
			unsigned char *pix = img.data+img.step[0]*y+x*channels;
			*((unsigned short*)(bucketImg.data+bucketImg.step[0]*y)+x) =
				(*(pix+idxR) >> hist_tobin) * hist_size2 +
				(*(pix+idxG) >> hist_tobin) * hist_size  +
				(*(pix+idxB) >> hist_tobin);
		}
		
		return bucketImg;
	}


} // namespace histodet

