#include <iostream>
#include <iomanip>
#include <typeinfo>

#include <opencv2/highgui/highgui.hpp>

#include "histodet/histodet.hpp"
#include "histodet/histogramBruteforce.hpp"
#include "histodet/labelReader.hpp"
#include "histodet/refinerCorrelation.hpp"
#include "histodet/refinerHueBlobs.hpp"

#if defined SSE && !defined __SSE__
	#warning "Argh, cannot use SSE1 optimizations on this platform!"
#endif

using namespace histodet;

// fine setup
//typedef DistanceL1n Distance;
typedef DistancePOSn Distance;

typedef BucketerHueLowsat<true, true> Bucketer;
//typedef BucketerHueLowsat<false, false> Bucketer;
//typedef BucketerHueValue<false> Bucketer;

typedef DetectorHistogram<HistogramDistributive<char>, Distance> DH;
//typedef ObjectExtractorBiggest ObjectExtractor;
//typedef ObjectExtractorKmeans ObjectExtractor;
typedef ObjectExtractorFloodFill ObjectExtractor;
//typedef RefinerCorrelation ObjectRefiner;
typedef RefinerHueBlobs ObjectRefiner;
const bool do_stretch = false;
const bool do_stretch_normalize = false;

const int nthres = 101;
const int nqual = 2;
const int nthres_ref = 21;
const float thres_ref_max = 20.0;


bool checkRect(cv::Rect const& rect, int w, int h)
{
	return !(rect.x < 0 || rect.y < 0 || rect.x+rect.width > w || rect.y+rect.height > h);
}

void learn(char *histv, int nimages, char** imgv, char *maskv, int max_qual = 0, int hist_nbits = 4)
{
 	BucketerAbstract *bucketer = NULL;
	if (typeid(Bucketer) == typeid(BucketerGray))
		bucketer = new BucketerGray(hist_nbits);
	else if (typeid(Bucketer) == typeid(BucketerHue<false>))
		bucketer = new BucketerHue<false>(1<<hist_nbits, hist_nbits+1);
	else if (typeid(Bucketer) == typeid(BucketerHue<true>))
		bucketer = new BucketerHue<true>(1<<hist_nbits, hist_nbits+1);
	else if (typeid(Bucketer) == typeid(BucketerHueBruteforce))
		bucketer = new BucketerHueBruteforce(hist_nbits);
	else if (typeid(Bucketer) == typeid(BucketerHueLowsat<true,false>))
		bucketer = new BucketerHueLowsat<true,false>(1<<hist_nbits, hist_nbits+1, 64, 32);
	else if (typeid(Bucketer) == typeid(BucketerHueLowsat<true,true>))
		bucketer = new BucketerHueLowsat<true,true>(1<<hist_nbits, hist_nbits+1, 64, 32);
	else if (typeid(Bucketer) == typeid(BucketerHueLowsat<false,false>))
		bucketer = new BucketerHueLowsat<false,false>(1<<hist_nbits, hist_nbits+1, 64, 32);
	else if (typeid(Bucketer) == typeid(BucketerHueLowsat<false,true>))
		bucketer = new BucketerHueLowsat<false,true>(1<<hist_nbits, hist_nbits+1, 64, 32);
	else if (typeid(Bucketer) == typeid(BucketerHueValue<false>))
		bucketer = new BucketerHueValue<false>(1<<(hist_nbits-1), 1<<(hist_nbits-1), hist_nbits);
	else if (typeid(Bucketer) == typeid(BucketerHueValue<true>))
		bucketer = new BucketerHueValue<true>(1<<(hist_nbits-1), 1<<(hist_nbits-1), hist_nbits);
	else if (typeid(Bucketer) == typeid(BucketerRGB))
		bucketer = new BucketerRGB(hist_nbits);
	
	DH det(hist_nbits, bucketer);
	RefinerHueBlobs refiner;
	det.configureStretching(do_stretch, do_stretch_normalize);
	LabelReader label(true);
	cv::Mat img;
	cv::Rect rect;
	int status, quality;

	det.learnBegin(maskv);
	
	for(int k = 0; k < nimages; ++k)
	{
// std::cout << "Learning image " << imgv[k] << std::endl;
		if (!label.loadBoundingBox(std::string(imgv[k])+".txt", rect, status, quality))
			{ std::cout << "no bounding box label file for " << imgv[k] << std::endl; continue; }
		if (status == 0 || status == 3 || quality > max_qual) continue;

		if (!(img = cv::imread(imgv[k], -1)).data)
			{ std::cout << "could not load " << imgv[k] << std::endl; continue; }
		if (!checkRect(rect, img.cols, img.rows))
			{ std::cout << "object too close to border in " << imgv[k] << std::endl; continue; }

		double size, ori, center_x, center_y;
		if (!label.loadSizeOri(std::string(imgv[k])+".so.txt", size, ori, center_x, center_y))
			{ std::cout << "no sizeOri label file for " << imgv[k] << std::endl; continue; }
		std::vector<std::pair<double,double> > blobs;
		if (!label.loadBlobs(std::string(imgv[k])+".b.txt", blobs))
			{ std::cout << "no blob label file for " << imgv[k] << std::endl; continue; }

		det.learnImage(img, rect, size, ori, center_x, center_y);

		if (k == 0) refiner.learnBegin(blobs.size());
		refiner.learnImage(img, rect, size, ori, center_x, center_y, blobs);
	}

	std::ostringstream name; name << histv << "_" << max_qual;
	det.learnEnd(name.str());
	refiner.learnEnd("blob");

	delete bucketer;
}

void detect(char *histv, char *correlv, int nimages, char** imgv, int detect_size, double detect_thres, int detect_margin = 1)
{
	int hist_size, hist_nbits;
 	BucketerAbstract *bucketer = BucketerAbstract::createBucketer(histv, hist_size, hist_nbits);
	DH det(hist_nbits, bucketer);
	ObjectExtractor extractor;
	//RefinerCorrelation refiner(correlv);
	RefinerHueBlobs refiner;
	refiner.refineInit(correlv);
	
	LabelReader label(true);
	cv::Mat img;
	cv::Rect rect;
	int status, quality;
	int falsep_margin = detect_size;
	bool dist_inverted;
	int ndetect_margin = sqr(detect_margin*2+1);

	struct DetRes
	{
		double thres;
		double thres_norm;
		long long int nfalsep[2];
		long long int nfound[2*nqual];
		
		double thres_ref[nthres_ref];
		int nfalsep_ref[nthres_ref];
		int nfound_ref[nthres_ref*nqual];
		
		int found[nqual];
		DetRes()
		{
			for(int i = 0; i < 3*nqual; ++i) nfound[i] = 0;
			nfalsep[0] = nfalsep[1] = 0;
			for(int r = 0; r < nthres_ref; ++r) thres_ref[r] = r*thres_ref_max/(nthres_ref-1);
			for(int r = 0; r < nthres_ref; ++r) nfalsep_ref[r] = 0;
			for(int r = 0; r < nthres_ref*nqual; ++r) nfound_ref[r] = 0;
		}
	};
	DetRes res[nthres];
	for(int n = 0; n < nthres; ++n) res[n].thres_norm = n*(1.0/(nthres-1));
	long long int npix = 0;
	int nobjects[nqual] = {0};
	
	if (!det.detectInit(histv)) return;
	
	for(int n = 0; n < nthres; ++n)
		res[n].thres = Distance::transform_threshold(res[n].thres_norm, detect_size*detect_size, hist_size, dist_inverted);
	double detect_thres_raw = Distance::transform_threshold(detect_thres, detect_size*detect_size, hist_size, dist_inverted);

	bool global_stats = (nimages > 10);
	
	for(int k = 0; k < nimages; ++k)
	{
		std::cout << "." << std::flush;

		bool stats = true;
		if (!label.loadBoundingBox(std::string(imgv[k])+".txt", rect, status, quality))
			{ if (global_stats) std::cout << "no label file for " << imgv[k] << ", this image will be ignored for statistics" << std::endl; stats = false; }
		if (status == 0) stats = false;

		if (!(img = cv::imread(imgv[k], -1)).data)
			{ std::cout << "could not load " << imgv[k] << std::endl; continue; }
		rect = cv::Rect(rect.x+rect.width/2-detect_size/2, rect.y+rect.height/2-detect_size/2, detect_size, detect_size);
		if (!checkRect(rect, img.cols, img.rows))
			quality = nqual;

		cv::Mat res_img = det.detectImage(img, detect_size);
		double score;

		if (global_stats && stats)
		{
			cv::Point center(rect.x+rect.width/2, rect.y+rect.height/2);

			// compute detection rate at pixel level
			/* An object is considered as detected if one pixel within a distance of detect_margin from
				the label center is above the threshold. */
			if (status != 3) for(int m = 0; m < nqual; ++m) if (quality == m)
			{
				nobjects[m]++;
				for(int n = 0; n < nthres; ++n) res[n].found[m] = 0;

				for(int i = center.y-detect_margin; i <= center.y+detect_margin; ++i)
					for(int j = center.x-detect_margin; j <= center.x+detect_margin; ++j)
						if (i >= detect_size/2 && i < res_img.rows-detect_size/2 &&
								j >= detect_size/2 && j < res_img.cols-detect_size/2)
							for(int n = 0; n < nthres; ++n)
								if (Distance::compare(res_img.at<float>(i,j), res[n].thres)) res[n].found[m]++;

				for(int n = 0; n < nthres; ++n)
					if (res[n].found[m]) res[n].nfound[m]++;
			}


			// compute false positive rate at pixel level
			/* A pixel is considered as false positive if it is above the threshold
				and further from the label center than falsep_margin. */
			for(int i = detect_size/2; i < res_img.rows-detect_size/2; ++i)
				for(int j = detect_size/2; j < res_img.cols-detect_size/2; ++j)
					if (status == 3 ||
							!(i >= center.y-falsep_margin && i <= center.y+falsep_margin &&
								j >= center.x-falsep_margin && j <= center.x+falsep_margin))
					{
						npix++;
						for(int n = 0; n < nthres; ++n)
							if (Distance::compare(res_img.at<float>(i,j), res[n].thres)) res[n].nfalsep[0]++;
					}

			// compute detection and false positive rate at object level
			/* an object is considered as detected if the extractor
				returned a box whose center is closer than winsize/2 and smaller than winsize*2 */
			for(int n = 0; n < nthres; ++n)
			{
				bool output = (detect_thres > 0 && fabs(detect_thres_raw - res[n].thres) < (0.5/(nthres-1)));
				
				ObjectList objs = extractor.extract(res_img, res[n].thres, dist_inverted, detect_size);
				bool missed = (status != 3 && quality < nqual);
				bool detected = false;
				for(ObjectList::iterator it = objs.begin(); it != objs.end(); ++it)
				{
					Object & obj = *it;
					bool confirmed = refiner.refine(obj, img, score);

					// detection
					if (status != 3 && quality < nqual && (std::abs(center.x-obj.loc[Object::lpPosU]) <= detect_size/2 && std::abs(center.y-obj.loc[Object::lpPosV]) <= detect_size/2) && !detected)
					{
						res[n].nfound[nqual+quality]++;
						for(int r = nthres_ref-1; r >= 0; --r) if (score < res[n].thres_ref[r]) res[n].nfound_ref[nthres_ref*quality+r]++; else break;
						if (confirmed) missed = false;
						detected = true;
					}
					// false positive
					if (status == 3 || (std::abs(center.x-obj.loc[Object::lpPosU]) > detect_size || std::abs(center.y-obj.loc[Object::lpPosV]) > detect_size))
					{
						res[n].nfalsep[1]++;
						for(int r = nthres_ref-1; r >= 0; --r) if (score < res[n].thres_ref[r]) res[n].nfalsep_ref[r]++; else break;
						if (confirmed && output) std::cout << "False positive: " << imgv[k] << std::endl;
					}
				}
				if (missed && output) std::cout << "Missed: " << imgv[k] << std::endl;
			}
		}
		
		// write image results
		if (detect_thres > 0)
		{
			cv::Mat img_gray, img_write, img_mask, img_dark, img_channels[3], res_img_norm, res_img_thr;
			res_img_norm = cv::Mat::zeros(res_img.rows, res_img.cols, CV_32F);
			for(int i = 0; i < res_img.rows; ++i) for(int j = 0; j < res_img.cols; ++j)
				res_img_norm.at<float>(i,j) = Distance::normalize(res_img.at<float>(i,j), detect_size*detect_size, hist_size);
			res_img_norm.convertTo(img_write, CV_8U, 255., 0.0);
			cv::imwrite(std::string(imgv[k])+".detect1.jpg", img_write);
			
			if (img.channels() >= 3) cvtColor(img, img_gray, CV_BGR2GRAY); else img_gray = img;
			cv::threshold(res_img_norm, res_img_thr, detect_thres, 1.0, cv::THRESH_BINARY_INV);
			cv::threshold(img_gray, img_dark, 128, 1, cv::THRESH_BINARY);
			for(int i = 0; i < 3; ++i) img_gray.copyTo(img_channels[i]);
			res_img_thr.convertTo(img_mask, CV_8U, 255., 0.0);
			img_channels[0].setTo(255, img_mask.mul(img_dark));
			img_channels[1].setTo(0, img_mask.mul(255-img_dark));
			cv::merge(img_channels, 3, img_write);
			
			ObjectList objs = extractor.extract(res_img, detect_thres_raw, dist_inverted, detect_size);
			for(ObjectList::iterator it = objs.begin(); it != objs.end(); ++it)
			{
				Object & obj = *it;
				cv::rectangle(img_write, cv::Rect(obj.loc[Object::lpPosU]-detect_size/2, obj.loc[Object::lpPosV]-detect_size/2, detect_size, detect_size), cv::Scalar(0,255,0), 1);
				if (refiner.refine(obj, img, score))
				{
// std::cout << "Refiner FOUND !!" << std::endl;
// std::cout << "pos " << obj.loc[Object::lpPosU] << "," << obj.loc[Object::lpPosV] << " ; ori " << obj.loc[Object::lpHeading] << " ; size " << obj.loc[Object::lpSize] << std::endl;
					cv::line(img_write,
						cv::Point(
							obj.loc[Object::lpPosU]+cos(obj.loc[Object::lpHeading])*obj.loc[Object::lpSize]/2,
							obj.loc[Object::lpPosV]-sin(obj.loc[Object::lpHeading])*obj.loc[Object::lpSize]/2),
						cv::Point(
							obj.loc[Object::lpPosU]-cos(obj.loc[Object::lpHeading])*obj.loc[Object::lpSize]/2,
							obj.loc[Object::lpPosV]+sin(obj.loc[Object::lpHeading])*obj.loc[Object::lpSize]/2),
						cv::Scalar(0,255,0), 2);
					cv::line(img_write,
						cv::Point(
							obj.loc[Object::lpPosU]+cos(obj.loc[Object::lpHeading])*obj.loc[Object::lpSize]/2-cos(obj.loc[Object::lpHeading]+M_PI/2)*5,
							obj.loc[Object::lpPosV]-sin(obj.loc[Object::lpHeading])*obj.loc[Object::lpSize]/2+sin(obj.loc[Object::lpHeading]+M_PI/2)*5),
						cv::Point(
							obj.loc[Object::lpPosU]+cos(obj.loc[Object::lpHeading])*obj.loc[Object::lpSize]/2+cos(obj.loc[Object::lpHeading]+M_PI/2)*5,
							obj.loc[Object::lpPosV]-sin(obj.loc[Object::lpHeading])*obj.loc[Object::lpSize]/2-sin(obj.loc[Object::lpHeading]+M_PI/2)*5),
						cv::Scalar(0,255,0), 2);
				}
			}
			
			cv::imwrite(std::string(imgv[k])+".detect2.jpg", img_write);
		}
	}

	// write stats file
	if (global_stats)
	{
		std::ofstream file;
		std::ostringstream name; name << "res_" << detect_margin << "_" << detect_size << ".dat";
		file.open(name.str().c_str());

		file << "# hist [" << histv << "] detect_margin " << detect_margin << " detect_size " << detect_size << std::endl;
		file << "# each successive group of detection rate corresponds to increasing levels of difficulty (decreasing quality)" << std::endl;
		file << "# number of images without obj : " << nimages << std::endl;
		for(int m = 0; m < nqual; ++m) file << "# number of images of qual " << m << " : " << nobjects[m] << std::endl;
		file << "# thres_hist\tthres_ref\tfalsep_pix_rate0\tfalsep_obj_rate0\tfalsep_ref_rate0\tdet_pix_rate0\tdet_obj_rate0\tdet_ref_rate0[...]" << std::endl;
		file << std::setfill(' ');
		for(int n = 0; n < nthres; ++n)
		{
			for(int r = 0; r < nthres_ref; ++r)
			{
				file << std::setw(6) << res[n].thres_norm << " " << std::setw(6) << res[n].thres_ref[r];
				file << " " << std::setw(11) << res[n].nfalsep[0]    /(double)npix*100
				     << " " << std::setw(11) << res[n].nfalsep[1]    /(double)nimages*100
				     << " " << std::setw(11) << res[n].nfalsep_ref[r]/(double)nimages*100;
				for(int m = 0; m < nqual; ++m)
					file << " " << std::setw(11) << res[n].nfound[m]                 /(double)nobjects[m]*100
					     << " " << std::setw(11) << res[n].nfound[nqual+m]           /(double)nobjects[m]*100
					     << " " << std::setw(11) << res[n].nfound_ref[nthres_ref*m+r]/(double)nobjects[m]*100;
				file << std::endl;
			}
			file << "#" << std::endl;
		}
		file.close();
	}

	delete bucketer;
}


int main(int argc, char** argv)
{
	if (argc < 2) { std::cout << "no command provided (choose in {learn,detect})" << std::endl; return 1; }
	std::string command = argv[1];
	if (command == "learn")
	{
		int narg = 4;
		/**
		 @param hist_prefix prefix of the generated histogram file name
		 @param mask mask description file if want to do background removal
		 @param max_qual maximum quality index (ie minimum quality) of patches used for learning (0 = only good, 1 = ok or good)
		 @param hist_nbits number of bits of the histogram generated, ie the histogram will have (1<<hist_nbits) bins
		 */
		if (argc < narg+3) { std::cout << "Usage : histodet_demo learn <hist_prefix> <mask> <max_qual> <hist_nbits> <...images...>" << std::endl; return 1; }
		learn(argv[2+0], argc-narg-2, argv+narg+2, argv[2+1], atoi(argv[2+2]), atoi(argv[2+3]));
	} else if (command == "detect")
	{
		int narg = 5;
		/**
		 @param hist_file histogram xml file name to detect
		 @param detect_size detection patch size
		 @param detect_thres threshold of detection for image result outputs. If negative, no image result output is written.
		 @param detect_margin distance to the center of the labeled position that is considered as a detection of the object for the detection rate
		 */
		if (argc < narg+3) { std::cout << "Usage : histodet_demo detect <hist_file> <refiner_file> <detect_size> <detect_thres> <detect_margin> <...images...>" << std::endl; return 1; }
		detect(argv[2+0], argv[2+1], argc-narg-2, argv+narg+2, atoi(argv[2+2]), atof(argv[2+3]), atoi(argv[2+4]));
	} else
	{ std::cout << "unknown command " << argv[1] << " (choose in {learn,detect})" << std::endl; return 1; }

	return 0;
}

