
#include <opencv2/imgproc/imgproc.hpp>

#include "histodet/objectExtractor.hpp"


namespace histodet {


void ObjectExtractorAbstract::extractPoints(cv::Mat const & detImage, float threshold, bool inverted, int window_size)
{
	unsigned width = detImage.cols, height = detImage.rows;
	points = cv::Mat(0, 2, CV_32F);
	if (inverted)
	{
		for(unsigned y = 0; y < height; ++y)
			for(unsigned x = 0; x < width; ++x)
				if (detImage.at<float>(y,x) >= threshold)
					points.push_back(cv::Point_<float>(x,y));
	} else
	{
		for(unsigned y = 0; y < height; ++y)
			for(unsigned x = 0; x < width; ++x)
				if (detImage.at<float>(y,x) <= threshold)
					points.push_back(cv::Point_<float>(x,y));
	}
}




ObjectList ObjectExtractorBiggest::extract(cv::Mat const & detImage, float threshold, bool inverted, int window_size)
{
#ifdef CHRONO
	static struct timeval tv1, tv2;
	static double avgt = 0., mint = 1e6;
	static int nchrono = 0;
	gettimeofday(&tv1,NULL);
#endif

	cv::Rect rect(0, 0, detImage.cols, detImage.rows);
	unsigned min_size = ((window_size * 3/2) / 2)*2+1; // ensure odd
	unsigned prev_pixels = 0;
	bool converged = false;
	unsigned pixels, sum_x, sum_y;

	extractPoints(detImage, threshold, inverted, window_size);

	while (!converged)
	{
		// compute new barycenter
		pixels = sum_x = sum_y = 0;
		for(size_t i = 0; i < points.rows; ++i)
		{
			cv::Point p(points.at<float>(i,0), points.at<float>(i,1));
			if (isInRect(p, rect)) { ++pixels; sum_x += p.x; sum_y += p.y; }
		}

		// checks
		if (pixels < 1 || pixels < prev_pixels/2) break; // too many false positives or not only one object
		sum_x /= pixels; sum_y /= pixels;
		if (rect.width == min_size && rect.height == min_size) { converged = true; break; }

		// iterate
		unsigned size = std::min(rect.width, rect.height) / 2;
		if (size < min_size) size = min_size;
		rect = cv::Rect(sum_x-size/2, sum_y-size/2, size, size);
		prev_pixels = pixels;
	}

	ObjectList result;
	int min_points = window_size*window_size/4; // FIXME set this in histo file?
	if (converged && pixels > min_points)
	{
		// TODO check the cluster is not actually bigger

		Object obj(1);
		obj.set(Object::lpPosU, sum_x, window_size / 3.0 / 3.0);
		obj.set(Object::lpPosV, sum_y, window_size / 3.0 / 3.0);
		obj.set(Object::lpSize, window_size, window_size / 3.0 / 3.0);
		result.push_back(obj);
	}

#ifdef CHRONO
	gettimeofday(&tv2,NULL);
	double t = (tv2.tv_sec-tv1.tv_sec) + (tv2.tv_usec-tv1.tv_usec)*1e-6;
	avgt += t; if (t < mint) mint = t; ++nchrono;
	std::cout << "[ObjectExtractor] Processed image in " << t << " seconds (average " << avgt/nchrono << " , min " << mint << ")" << std::endl;
#endif

	return result;
}



	struct Cluster
	{
		public:
			typedef cv::Point_<double> Point;
			typedef std::vector<Point> PointList;
			PointList points;
			mutable double x, y;
			mutable double xx, yy, xy;
			unsigned n;
			bool enabled;

		private:
			double sum_x, sum_y;
			double sum_xx, sum_yy, sum_xy;
			mutable bool up_to_date;

		public:
			Cluster(): sum_x(0.), sum_y(0.), sum_xx(0.), sum_yy(0.), sum_xy(0.), n(0), enabled(true), up_to_date(false) {}
			
			void addPoint(Point p)
			{
				points.push_back(p);
				++n;
				sum_x += p.x;
				sum_y += p.y;
				sum_xx += p.x*p.x;
				sum_yy += p.y*p.y;
				sum_xy += p.x*p.y;
				up_to_date = false;
			}

			
			void removeOutliers(double nsigma, int nsteps)
			{
				bool has_erased = true;
				for(int k = 0; k < nsteps && has_erased; ++k)
				{
					compute();
					has_erased = false;
					bool erase = false;
					for (unsigned i = 0; i < n; i=i+(erase?0:1))
					{
						Point &p = points[i];
						if (mahalanobis(p) > nsigma)
						{
							--n;
							sum_x -= p.x;
							sum_y -= p.y;
							sum_xx -= p.x*p.x;
							sum_yy -= p.y*p.y;
							sum_xy -= p.x*p.y;
							
							points[i] = points[n];
							points.resize(n);
							erase = true;
							has_erased = true;
						} else erase = false;
					}
				}
			}
			
			void compute() const
			{
				if (!up_to_date)
				{
					x = sum_x/n;
					y = sum_y/n;
					xx = sum_xx/n - x*x;
					yy = sum_yy/n - y*y;
					xy = sum_xy/n - x*y;
					up_to_date = true;
				}
			}

			double distance(const Cluster& cluster) const
			{
				compute();
				cluster.compute();
				//return std::min(mahalanobis(Point(cluster.x,cluster.y)), cluster.mahalanobis(Point(x,y)));
				
				Point q(cluster.x-x, cluster.y-y);
				double mxx = (xx+cluster.xx), mxy = (xy+cluster.xy), myy = (yy+cluster.yy);
				return (mxx*q.y*q.y - 2*mxy*q.x*q.y + myy*q.x*q.x) / (mxx*myy-mxy*mxy);
			}

		private:
			double mahalanobis(Point p) const
			{
				Point q(p.x-x, p.y-y);
				return (xx*q.y*q.y - 2*xy*q.x*q.y + yy*q.x*q.x) / (xx*yy-xy*xy);
			}

	};


ObjectList ObjectExtractorKmeans::extract(cv::Mat const & detImage, float threshold, bool inverted, int window_size)
{
#ifdef CHRONO
	static struct timeval tv1, tv2;
	static double avgt = 0., mint = 1e6;
	static int nchrono = 0;
	gettimeofday(&tv1,NULL);
#endif
#if DISPLAY
	const int nsteps = 3;
	const int ncolors = nsteps*nsteps*nsteps - nsteps;
	static const uchar steps[nsteps] = {0,128,255};
	uchar colors[ncolors][3];// = {{255,0,0},{0,255,0},{0,0,255},{0,255,255},{255,0,255},{255,255,0},{128,0,0},{0,128,0},{0,0,128},{0,128,128};
	for(int i = 0, c = 0; i < 3; ++i) for(int j = 0; j < 3; ++j) for(int k = 0; k < 3; ++k)
	{
		if (i == j && j == k) continue;
		colors[c][0] = steps[i];
		colors[c][1] = steps[j];
		colors[c][2] = steps[k];
		++c;
	}
#endif
	// remove noise ? not sure... it would require processing the whole image, and can be done by removing tiny clusters or points far away from their cluster
	if (0)
	{
		cv::Mat cleanImage = cv::Mat(detImage.rows, detImage.cols, CV_32F, cv::Scalar(0));
		const int denoise_size = 1;
		cv::Mat element = getStructuringElement(cv::MORPH_CROSS, cv::Size(2*denoise_size+1, 2*denoise_size+1), cv::Point(denoise_size, denoise_size));
		erode(detImage, cleanImage, element, cv::Point(-1, -1), 1);
		dilate(cleanImage, cleanImage, element, cv::Point(-1, -1), 1);
	}
	
	extractPoints(detImage, threshold, inverted, window_size);

	// kmeans
	cv::Mat labels;
	int nclusters = std::min(10, points.rows/3); // we want clusters with at least 3 pixels
	std::vector<Cluster> clusters;
	clusters.resize(nclusters);
	
	cv::kmeans(points, nclusters, labels, cv::TermCriteria(cv::TermCriteria::COUNT, 10, 0.0), 1, cv::KMEANS_PP_CENTERS);

	for(int k = 0; k < labels.rows; ++k)
		clusters[labels.at<int>(k)].addPoint(points.at<cv::Point_<float> >(k));

#if DISPLAY
	static int nim = 0;
	std::ostringstream oss; oss << std::setw(2) << std::setfill('0') << nim;
	cv::Mat imclusters = cv::Mat(detImage.rows, detImage.cols, CV_8UC3, cv::Scalar(0,0,0));
	for(int i = 0; i < nclusters; ++i) if (clusters[i].enabled)
		for(int k = 0; k < clusters[i].n; ++k) for(int j = 0; j < 3; ++j)
			imclusters.at<uchar[3]>(clusters[i].points[k].y,clusters[i].points[k].x)[j] = colors[i][j];
	cv::imwrite(oss.str()+std::string("_clusters1.png"), imclusters);
#endif
	
	// remove noise
	for(int i = 0; i < nclusters; ++i)
		clusters[i].removeOutliers(2.5*2.5, 3);

#if DISPLAY
	imclusters = cv::Mat(detImage.rows, detImage.cols, CV_8UC3, cv::Scalar(0,0,0));
	for(int i = 0; i < nclusters; ++i) if (clusters[i].enabled)
		for(int k = 0; k < clusters[i].n; ++k) for(int j = 0; j < 3; ++j)
			imclusters.at<uchar[3]>(clusters[i].points[k].y,clusters[i].points[k].x)[j] = colors[i][j];
	cv::imwrite(oss.str()+std::string("_clusters2.png"), imclusters);
#endif
	
	// merge colliding clusters
	int n = nclusters;
	bool merged_clusters = true;
	while (merged_clusters)
	{
		merged_clusters = false;
		for(int i = 0; i < nclusters; ++i)
		{
			if (clusters[i].enabled)
			for(int j = i+1; j < nclusters; ++j)
			{
				if (clusters[j].enabled)
				{
				if (clusters[i].distance(clusters[j]) < 3.*3.)
				{
					for(int k = 0; k < clusters[j].n; ++k)
						clusters[i].addPoint(clusters[j].points[k]);
					clusters[i].compute();
					clusters[j].enabled = false;
					--n;
					merged_clusters = true;
				}
				}
			}
		}
	}

#if DISPLAY
	imclusters = cv::Mat(detImage.rows, detImage.cols, CV_8UC3, cv::Scalar(0,0,0));
	for(int i = 0; i < nclusters; ++i) if (clusters[i].enabled)
		for(int k = 0; k < clusters[i].n; ++k) for(int j = 0; j < 3; ++j)
			imclusters.at<uchar[3]>(clusters[i].points[k].y,clusters[i].points[k].x)[j] = colors[i][j];
	cv::imwrite(oss.str()+std::string("_clusters3.png"), imclusters);
#endif

	
	ObjectList result;
	int min_points = window_size*window_size/4; // FIXME set this in histo file?
	int max_points = window_size*window_size*6;

	// TODO extract long axis of ellipse to verify it is not too long wrt window_size
	
	for(int i = 0; i < nclusters; ++i)
		if (clusters[i].enabled && clusters[i].n > min_points && clusters[i].n < max_points)
		{
			Object obj(1);
			obj.set(Object::lpPosU, clusters[i].x, window_size / 3.0 / 3.0);
			obj.set(Object::lpPosV, clusters[i].y, window_size / 3.0 / 3.0);
			obj.set(Object::lpSize, window_size, window_size / 3.0 / 3.0);
			result.push_back(obj);
		}

#ifdef CHRONO
	gettimeofday(&tv2,NULL);
	double t = (tv2.tv_sec-tv1.tv_sec) + (tv2.tv_usec-tv1.tv_usec)*1e-6;
	avgt += t; if (t < mint) mint = t; ++nchrono;
	std::cout << "[ObjectExtractor] Processed image in " << t << " seconds (average " << avgt/nchrono << " , min " << mint << ")" << std::endl;
#endif

	return result;
}




} // namespace histodet


